#include <ctype.h> // For isprint()
#include <libc.h>
#include <stdio.h>
#include <ndbm.h>
#include "../server/database.c"

static void CheckUids(void);
static void RepairUser(bb_uid uid);
static void CheckRids(void);
static void CheckRoomPointers(void);
static const char *FindNameForRoom(bb_rid rid);
static void ListDuplicateRoomNames(bb_rid rid, const char *roomName);
static void RepairRoom(bb_rid rid);
static const char *nameForClass(class_t class);
static int validateClass(class_t class);
static int validateRoomName(const char *string);
static int validateId(const char *string);

int repair_test_only = 0;

int main(int argc, char *argv[])
{
    char message_base[80] = MESSAGEBASE;
    int ch;
    extern char *optarg;
    int check_rooms = 1;

    setlinebuf(stdout);
    setlinebuf(stderr);
    while ((ch = getopt(argc, argv, "M:nu")) != EOF) {
	switch (ch) {
	    case 'M':
	        strcpy(message_base, optarg);
	        break;
            case 'n':
                repair_test_only = 1;
                printf("Test mode enabled\n\n");
                break;
            case 'u':
                check_rooms = 0;
                printf("Checking users only\n\n");
                break;
	}
    }

    if (chdir(message_base) != 0) {
	perror(message_base);
	return 1;
    }
    if (!database_init(repair_test_only, FALSE)) {
	perror("database init failed");
	return 1;
    }
    CheckUids();
    database_shutdown();
    if (check_rooms) {
        CheckRids();
        CheckRoomPointers();
    }
    return 0;
}

static void CheckUids(void)
{
    bb_uid uid;
    DbUser user_info;

    for (uid = 0; uid <= Last_uid; uid++) {
        if (!get_db_user(&user_info, uid))
            continue;
        RepairUser(uid);
    }
}

static void RepairUser(bb_uid uid)
{
    DbUser user_info;
    bb_uid uid_from_id;

    if (!get_db_user(&user_info, uid))
        return;
    if (!validateId(user_info.id)) {
        printf("*** User %d has an invalid id '%s'\n", (int)uid, user_info.id);
    }
    uid_from_id = get_uid(user_info.id);
    if (uid_from_id != uid) {
        printf("*** Repairing user %d: creating user id pointer '%s'\n", (int)uid, user_info.id);
        if (uid_from_id == -1) {
            if (!repair_test_only) {
                if (!put_uid(user_info.id, uid))
                    printf("*** Repairing user %d:   unable to create pointer!\n", (int)uid);
            }
        } else {
            printf("*** Repairing user %d:   unable to create id '%s', already used by %d\n", (int)uid, user_info.id, (int)uid_from_id);
        }
    }
}

static void CheckRids(void)
{
    bb_rid rid;

    if (!database_init(repair_test_only, FALSE)) {
	perror("database init failed");
	return;
    }
    for (rid = 0; rid <= Last_rid; rid++) {
        DbDirectory *db_dir;
        int count;

        /* load directory info */
        db_dir = create_db_directory(); /* constructor */
        
        /* room does not exist on disk */
        if (!get_db_directory(db_dir, rid)) {
            free(db_dir);
            continue;
        }

        count = db_dir->num_msgs;
        
        if (count < 0 || count > MAX_MESSAGES) {
            const char *roomName;
    
            roomName = FindNameForRoom(rid);
            printf("*** Repairing room %d ('%s'): message count overflow (%d), resetting to %d\n", (int)rid, roomName, count, MAX_MESSAGES);
            db_dir->num_msgs = MAX_MESSAGES;
            if (!repair_test_only) {
                if (!put_db_directory(rid, db_dir)) {
                    printf("*** Repairing room %d:   unable to update directory!\n", (int)rid);
                }
            }
        }
        RepairRoom(rid);
    }
    database_shutdown();
}

static void CheckRoomPointers(void)
{
    DBM *db;
    datum key;
    
#warning TO DO: Remove duplicate room name pointers
    return;

    db = dbm_open(ROOM, O_RDONLY, 0666);
    for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
        char roomNameBuffer[ROOM_SIZE];
	datum content;
        bb_rid rid;

        if (key.dsize != sizeof(rid_key) || key.dptr[0] != DATA_RID)
            continue;
	content = dbm_fetch(db, key);
	if (content.dptr == NULL) {
	    perror("dbm_fetch");
	    exit(1);
	}
        COPY(content.dptr, &rid, bb_rid);
        strcpy(roomNameBuffer, key.dptr + 1);
    }
    dbm_close(db);
}

static const char *FindNameForRoom(bb_rid rid)
{
    static char roomNameBuffer[ROOM_SIZE];
    DBM *db;
    datum key;

    *roomNameBuffer = '\0'; // Don't return stale information
    db = dbm_open(ROOM, O_RDONLY, 0666);
    for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
	datum content;
        bb_rid target_rid;

        if (key.dsize != sizeof(rid_key) || key.dptr[0] != DATA_RID)
            continue;
	content = dbm_fetch(db, key);
	if (content.dptr == NULL) {
	    perror("dbm_fetch");
	    exit(1);
	}
        COPY(content.dptr, &target_rid, bb_rid);
        if (target_rid == rid) {
            strcpy(roomNameBuffer, key.dptr + 1);
            break;
        }
    }
    dbm_close(db);
    return roomNameBuffer;
}

#ifdef DEBUG
static void ListDuplicateRoomNames(bb_rid rid, const char *roomName)
{
    DBM *db;
    datum key;
    bb_rid target_rid;

    db = dbm_open(ROOM, O_RDONLY, 0666);
    for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
	datum content;

        if (key.dsize != sizeof(rid_key) || key.dptr[0] != DATA_RID)
            continue;
	content = dbm_fetch(db, key);
	if (content.dptr == NULL) {
	    perror("dbm_fetch");
	    exit(1);
	}
        COPY(content.dptr, &target_rid, bb_rid);
        if (target_rid == rid && strcmp(key.dptr + 1, roomName) != 0)
            printf("\tduplicate room name at '%s'\n", key.dptr + 1);
    }
    dbm_close(db);
}
#endif

static void RepairRoom(bb_rid rid)
{
    DbDirectory *db_directory;

    db_directory = get_db_room_directory(rid);
    if (db_directory == NULL)
        return;

    if (!validateRoomName(db_directory->name)) {
        const char *roomName;

        roomName = FindNameForRoom(rid);
        if (roomName != NULL && !validateRoomName(roomName)) {
            printf("*** Repairing room %d: removing invalid name pointer '%s'\n", (int)rid, roomName);
            if (!repair_test_only)
                remove_rid(roomName);
            roomName = NULL;
        }
        if (roomName == NULL) {
            static char buffer[128];

            sprintf(buffer, "room-%d", (int)rid);
            roomName = buffer;
        }
        printf("*** Repairing room %d: name = '%s'\n", (int)rid, roomName);
        strcpy(db_directory->name, roomName);
    }
    if (!validateClass(db_directory->default_class)) {
        printf("*** Repairing room %d: default_class = no_access\n", (int)rid);
        db_directory->default_class = no_access;
    }
    if (!repair_test_only)
        put_db_room(rid);
    if (get_rid(db_directory->name) != rid) {
        printf("*** Repairing room %d: creating room name pointer '%s'\n", (int)rid, db_directory->name);
        if (!repair_test_only) {
            if (!put_rid(db_directory->name, rid))
                printf("*** Repairing room %d:   unable to create pointer!\n", (int)rid);
        }
    }
}

#ifdef DEBUG
static const char *nameForClass(class_t class)
{
    static char buffer[128];
    static char *classes[] = {"no_access", "read_only", "read_write", "moderator", "no_class"};

    switch (class) {
        case no_access:
        case read_only:
        case read_write:
        case moderator:
        case no_class:
            return classes[class];
        default:
            sprintf(buffer, "[%d]", class);
            return buffer;
    }
}
#endif

static int validateClass(class_t class)
{
    switch (class) {
        case no_access:
        case read_only:
        case read_write:
        case moderator:
        case no_class:
            return 1;
        default:
            return 0;
    }
}

static int validateRoomName(const char *string)
{
    const char *s;

    if (*string == '\0')
        return 0; // Empty string
    for (s = string; *s != '\0'; s++)
	if (!isprint(*s) && *s != '\t')
	    return 0; // Invalid character
    return 1;
}

static int validateId(const char *string)
{
    const char *s;

    if (!*string)
	return 0; // Empty string
    for (s = string; *s; s++)
	if ((!isprint(*s)) || (*s == ' ') || (*s == '/'))
	    return 0; // Invalid character
    return 1;
}
